import logging
import sys
import os
import json

from voluptuous import Schema

from confgetti.base import get_variables


log = logging.getLogger(__name__)


def set_values(config_module_name, values):
    """
    Set given values to config module.

    :param config_module: python module to set values to
    :type config_module: module
    :param values: keys and values to set
    :type values: dictionary
    """
    config_module = sys.modules[config_module_name]
    for key, value in values.items():
        setattr(config_module, key, value)


def dict_keys_to_uppercase(dict_for_convert):
    """
    Converts keys of provided dict to uppercase with same value

    :param dict_for_convert: dictionary whose keys will be uppercased
    :type dict_for_convert: dictionary

    :returns: dictionary with uppercase keys
    :rtype: dictionary
    """
    converted_dict = {}

    for key, value in dict_for_convert.items():
        converted_dict[key.upper()] = value

    return converted_dict


def load_from_json(env_var):
    """
    Load config from json file.

    :param env_var: name of the env var containing path to json file.
    :type env_var: string

    :returns: config
    :rtype: dictionary
    """
    path_to_json = os.environ.get(env_var)
    if path_to_json is None:
        log.warning("Config path set to None, unable to load "
                    "configuration: {}".format(env_var))
        return {}

    if not os.path.isfile(path_to_json):
        log.warning("Config file does not exist, unable to load "
                    "configuration: {}".format(path_to_json))
        return {}

    with open(path_to_json, "r") as f:
        return json.load(f)


def load_from_env(env_prefix):
    """
    Load config from env variables.

    :param env_prefix: prefix of env var names to get values from.
    :type env_prefix: string

    :returns: config
    :rtype: dictionary
    """
    return {key[len(env_prefix) + 1:].lower(): value
            for key, value in os.environ.items()
            if key.startswith(env_prefix) and len(key) > len(env_prefix)}


def load_from_config_server(namespace, keys):
    """
    Loads configuration from configuration server.

    :param namespace: namespace under which app configuration is located.
    :type namespace: string
    :param keys: Set of keys for variables lookup.
    :type keys: dictionary/list
    """
    return get_variables(
        path=namespace,
        keys=keys,
        use_env=False,
        use_consul=True)


def load_and_validate_config(
        config_module_name,
        env_var,
        schema=None,
        keys=None,
        uppercase=False):
    """
    Load config, validate and set to given module.

    :param config_module_name: name of the python module to set config to.
    :type config_module_name: string
    :param env_var: name of the env var containing path to config file.
    :type env_var: string
    :param schema: schema to use for config validation.
    :type schema: voluptuous.Schema
    :param uppercase: should keys be returned as uppercase or no.
    :type uppercase: boolean
    """
    if keys is None and isinstance(schema, Schema):
        keys = list(schema.schema.keys())

    try:
        config = {}

        loaded_configs = [
            load_from_config_server(env_var, keys),
            load_from_json(env_var),
            load_from_env(env_var),
        ]

        for loaded in loaded_configs:
            if uppercase is True:
                loaded = dict_keys_to_uppercase(loaded)

            config.update(loaded)

        if schema is not None:
            config = schema(config)

        set_values(config_module_name, config)
    except:
        log.error("Config error", exc_info=True)
        raise



import logging


class DuplicateFilter(logging.Filter):
    """
    Logging filter for removing duplicate messages for stdout
    """
    def filter(self, record):
        """
        Compare incoming message with last log and if they are not same,
        return True which indicates that message should be displayed.

        :param record: message record
        :type record: string

        :returns: boolean indicating if message should be displayed or no
        :rtype: boolean
        """
        current_log = (record.module, record.levelno, record.msg)

        if current_log != getattr(self, "last_log", None):
            self.last_log = current_log
            return True

        return False




class UndefinedConnectionError(Exception):
    """
    Raise exception when connection to service is not defined
    """
    pass


class ConvertValueError(Exception):
    """
    Raise exception when coversion of value is not possible
    """
    pass



import os
import logging
import json

from requests.exceptions import ConnectionError

from confgetti.logger import DuplicateFilter
from confgetti.remote import ConsulInterface
from confgetti.exceptions import UndefinedConnectionError, ConvertValueError


log = logging.getLogger(__name__)
log.addFilter(DuplicateFilter())


class ValueConvert(object):
    def __init__(self):
        """
        Declares boolean comparison lists under which
        strings are converted to booleans.
        """
        self.false_compare_list = ['false', 'False']
        self.true_compare_list = ['true', 'True']

    def convert_bool(self, value):
        """
        Converts string to boolean if value found in one of compare lists.

        :param value: value for conversion
        :type value: string

        :returns: converted value to new type or in original type
        :rtype: boolean/string
        """
        if value in self.false_compare_list:
            value = False
        elif value in self.true_compare_list:
            value = True
        else:
            raise ConvertValueError

        return value

    def convert_int(self, value):
        """
        Converts string to integer.

        :param value: value for conversion
        :type value: string

        :returns: converted value to new type or in original type
        :rtype: integer/string
        """
        try:
            value = int(value)
        except ValueError:
            raise ConvertValueError

        return value

    def convert_float(self, value):
        """
        Converts string to float.

        :param value: value for conversion
        :type value: string

        :returns: converted value to new type or in original type
        :rtype: float/string
        """
        try:
            value = float(value)
        except ValueError:
            raise ConvertValueError

        return value

    def convert_str(self, value):
        """
        Converts value to string.

        :param value: value for conversion
        :type value: string

        :returns: converted value to new type or in original type
        :rtype: string
        """
        return str(value)

    def convert_dict(self, value):
        """
        Converts string to dictionary.

        :param value: value for conversion
        :type value: dictionary

        :returns: converted value to new type or in original type
        :rtype: dictionary/string
        """
        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError:
            raise ConvertValueError

        return value

    def convert(self, value, convert_to=None):
        """
        Seeks for existing method based on wanted type.
        Converts value from bytes to string.
        Runs convert method if one found, if method raises error it logs
        warning.
        If method does not exists it logs warning about not supported
        conversion try.

        :param value: value for conversion
        :type value: any
        :param value: name of wanted value type
        :type value: string

        :returns: converted value to new type or in original type
        :rtype: any
        """
        if type(value) == bytes:
            value = value.decode('ascii')

        if convert_to is not None:
            if isinstance(convert_to, str):
                convert_name = convert_to
            else:
                convert_name = convert_to.__name__

            convert_method_name = 'convert_{0}'.format(convert_name)
            has_convert_method = hasattr(self, convert_method_name)

            if has_convert_method is True:
                convert_method = getattr(self, convert_method_name)
                
                try:
                    value = convert_method(value)
                except ConvertValueError:
                    log.warning('"{0}" cannot be converted to {1}!'.format(
                        value, convert_name
                    ))
            else:
                log.warning(
                    'method for "{0}" does not exist!'.format(
                        convert_name
                    )
                )

        return value


class Confgetti(object):
    """
    Declares classes for easier override if custom logic is needed.
    """
    consul_interface_class = ConsulInterface
    value_convert_class = ValueConvert

    def __init__(self, consul_config=None, prepare_consul=True):
        """
        Uses passed consul configuration to initalize consul interface,
        if configuration is passed. In other case, uses default configuration
        which is defined from environment variables.

        :param prepare_consul: shoud consul client be prepared or no
        :type prepare_consul: boolean
        :param consul_config: dictionary holding consul configuration data
        :type consul_config: dictionary/None
        """
        self.prepare_consul = prepare_consul

        if consul_config is not None:
            self.consul = self.consul_interface_class()
            self.consul.create_connection(consul_config)
        else:
            self.consul = self.consul_interface_class(self.prepare_consul)

        self.value_convert = self.value_convert_class()

    def get_variable(
            self,
            key,
            path=None,
            fallback=None,
            convert_to=None,
            use_env=True,
            use_consul=True):
        """
        Gets variable by passed key.
        It gets variable from environment and assigns it for return.
        If variable is not found in environment it tries to get variable
        from Consul service.
        If variable is not found in environment or in Consul, it returns
        fallback value.
        If variable exists, it runs conversion method based on wanted type
        provided under 'convert_to' argument.

        :param key: key name of desired variable
        :type key: string
        :param path: location of variable on Consul storage.
        :type path: string/None
        :param fallback: value that is returned if variable value not found
        :type fallback: any
        :param use_env: Should method look into environment for variable or no
        :type use_env: boolean
        :param use_consul: Should method look into consul for variable or no
        :type use_consul: boolean

        :returns: variable value possibly from one source or fallback.
        :rtype: any
        """
        variable = None

        if use_env is True:
            variable = os.environ.get(key)

        if self.prepare_consul is True and use_consul is True \
                and variable is None:
            try:
                variable = self.consul.get_raw_value(key, path)
            except (ConnectionError, UndefinedConnectionError):
                log.warning('Not connected to consul on host '
                            '"{}". Please check your consul '
                            'connection parameters!'.format(
                                self.consul.connection.http.host
                            ))

        if variable is not None:
            variable = self.value_convert.convert(variable, convert_to)

        return variable if variable is not None else fallback

    def get_variables(
            self, path=None, keys=None, use_env=True, use_consul=True):
        """
        Gets multiple variables from environment or consul, based on path.
        Supports dict and list types `keys` argument.
        In case of dict, it uses key value as conversion type,
        in case of list it just passes each key to function for getting
        variable, and everything returned is string.

        :param path: location of variable on Consul storage.
        :type path: string/None
        :param keys: set of keys whose values should be returned back
        :type keys: dict/list
        :param use_env: Should method look into environment for variable or no
        :type use_env: boolean
        :param use_consul: Should method look into consul for variable or no
        :type use_consul: boolean

        :returns: dictionary including fetched variables.
        :rtype: dict
        """
        keys = [] if keys is None else keys
        variables = {}

        if isinstance(keys, dict) is True:
            for key, value in keys.items():
                variable = self.get_variable(
                    key=key,
                    path=path,
                    convert_to=value,
                    use_env=use_env,
                    use_consul=use_consul)

                if variable is not None:
                    variables[key] = variable
        elif isinstance(keys, list) is True:
            for key in keys:
                variable = self.get_variable(
                    key=key,
                    path=path,
                    use_env=use_env,
                    use_consul=use_consul)

                if variable is not None:
                    variables[key] = variable
        else:
            raise TypeError('"keys" argument should be list or dict')

        return variables


def get_variables(path, keys, use_env=True, use_consul=True):
    """
    Shorthand function for simple Confgetti setup that returns desired
    variables in dictionary.

    :param path: location of variable on Consul storage.
    :type path: string/None
    :param keys: set of keys whose values should be returned back
    :type keys: dict/list
    :param use_env: Should method look into environment for variable or no
    :type use_env: boolean
    :param use_consul: Should method look into consul for variable or no
    :type use_consul: boolean

    :returns: dictionary including fetched variables.
    :rtype: dict
    """
    cgtti = Confgetti()

    return cgtti.get_variables(path, keys, use_env, use_consul)



import os
import consul

from requests.exceptions import ConnectionError
from confgetti.exceptions import UndefinedConnectionError


class ConsulInterface(object):
    def __init__(self, prepare_connection=False):
        """
        Sets empty connection upon initialization.
        Constructs default consul configuration, that will be used in
        connection to running consul, from environment variables.
        If class object is initialized with 'prepare_connection' as True
        it runs method that connects to consul instance and attaches it to
        initialized object.

        :param prepare_connection: Initialization creates connection or not
        :type prepare_connection: boolean
        """
        self.connection = None
        self.default_consul_config = {
            'host': os.environ.get('CONSUL_HOST', 'consul'),
            'port': os.environ.get('CONSUL_PORT', 8500),
            'scheme': os.environ.get('CONSUL_SCHEME', 'http'),
            'token': os.environ.get('CONSUL_TOKEN'),
            'dc': os.environ.get('CONSUL_DC')
        }

        if prepare_connection is True:
            self.create_connection()

    def create_connection(self, config=None):
        """
        Creates connection to Consul service.
        Uses default consul configuration constructed from environment
        variables if alternate configuration is not passed to method.

        :param config: Initialization creates connection or not
        :type config: dictionary/None

        :returns: instance of consul client
        :rtype: consul.std.Consul object
        """
        config = self.default_consul_config if config is None else config

        if self.connection is None:
            self.connection = consul.Consul(**config)

        return self.connection

    def _check_connection(self):
        """
        In case when connection object is not defined on class instance,
        it raises error of undefined connection.
        """
        if self.connection is None:
            raise UndefinedConnectionError('Consul connection is not defined!')

    def get_raw_value(self, key, path=None):
        """
        Gets value from Consul's key value storage.
        Firstly, calls method for checking connection.
        Constructs key path if `path` is provided.
        Uses current connection to get value by key path.
        If storage returns data dict, it gets value from it.

        :param key: key for desired value
        :type key: string
        :param path: path where key is stored on Consul service
        :type path: string/None

        :returns: fetched value from Consul service.
        :rtype: bytes/None
        """
        self._check_connection()

        value = None
        key_path = key

        if path is not None:
            key_path = '{0}/{1}'.format(
                path, key
            )

        index, data = self.connection.kv.get(key_path)

        if data is not None:
            value = data.get('Value')

        return value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['confgetti']

package_data = \
{'': ['*']}

install_requires = \
['python-consul>=1.1.0,<2.0.0', 'voluptuous>=0.11.7,<0.12.0']

setup_kwargs = {
    'name': 'confgetti',
    'version': '0.1.11',
    'description': 'Package for getting configuration variables from remote configuration server, json files and environment variables in a simple way',
    'long_description': None,
    'author': 'Styria Digital Development',
    'author_email': 'development@styria.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
